import os

import numpy as np

import torch
from torch import nn
from sklearn import metrics


# A simple CNN model
class CNN(nn.Module):
    def __init__(self, number_of_classes, vocab_size, embedding_dim, input_len, device):
        super(CNN, self).__init__()

        self.device = device

        if number_of_classes == 2:
            self.is_multiclass = False
            number_of_output_neurons = 1
            loss = torch.nn.functional.binary_cross_entropy_with_logits
            output_activation = nn.Sigmoid()
        else:
            self.is_multiclass = True
            number_of_output_neurons = number_of_classes
            loss = torch.nn.CrossEntropyLoss()
            output_activation = lambda x: x

        self.embeddings = nn.Embedding(vocab_size, embedding_dim)
        self.cnn_model = nn.Sequential(
            nn.Conv1d(in_channels=embedding_dim, out_channels=16, kernel_size=8, bias=True),
            nn.BatchNorm1d(16),
            nn.ReLU(),
            nn.MaxPool1d(2),

            nn.Conv1d(in_channels=16, out_channels=8, kernel_size=8, bias=True),
            nn.BatchNorm1d(8),
            nn.MaxPool1d(2),

            nn.Conv1d(in_channels=8, out_channels=4, kernel_size=8, bias=True),
            nn.BatchNorm1d(4),
            nn.MaxPool1d(2),

            nn.Flatten()
        )
        self.dense_model = nn.Sequential(
            nn.Linear(self.count_flatten_size(input_len), 512),
            nn.Linear(512, number_of_output_neurons)
        )
        self.output_activation = output_activation
        self.loss = loss

    def count_flatten_size(self, input_len):
        zeros = torch.zeros([1, input_len], dtype=torch.long)
        x = self.embeddings(zeros)
        x = x.transpose(1, 2)
        x = self.cnn_model(x)
        return x.size()[1]

    def forward(self, x):
        x = self.embeddings(x)
        x = x.transpose(1, 2)
        x = self.cnn_model(x)
        x = self.dense_model(x)
        x = self.output_activation(x)
        return x

    def train_loop(self, dataloader, optimizer):

        train_loss, correct = 0, 0
        self.train()
        for x, y in dataloader:
            x, y = x.to(self.device), y.to(self.device)
            if self.is_multiclass:
                y = y[:, 0].long()

            # Compute prediction error
            pred = self(x)
            loss = self.loss(pred, y)

            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            train_loss += self.loss(pred, y).item()

            if self.is_multiclass:
                correct += (pred.argmax(1) == y).type(torch.float).sum().item()
            else:
                correct += (torch.round(pred) == y).sum().item()

        size = len(dataloader.dataset)
        num_batches = len(dataloader)

        train_loss /= num_batches
        correct /= size
        print(f"Train metrics: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {train_loss:>8f} \n")

    def fit(self, dataloader, epochs):
        optimizer = torch.optim.Adam(self.parameters())
        for t in range(epochs):
            print(f"Epoch {t}")
            self.train_loop(dataloader, optimizer)

    def test(self, dataloader):

        num_batches = len(dataloader)
        self.eval()

        all_predictions = []
        all_labels = []
        test_loss = 0

        with torch.no_grad():
            for X, y in dataloader:
                if self.is_multiclass:
                    y = y[:, 0].long()

                output = self(X)
                test_loss += self.loss(output, y).item()

                if self.is_multiclass:
                    pred = output.argmax(1)
                else:
                    pred = torch.round(output)

                all_labels.extend(y.cpu().numpy())
                all_predictions.extend(pred.cpu().numpy())

        average = 'micro' if self.is_multiclass else 'binary'
        acc = metrics.accuracy_score(all_labels, all_predictions)
        f1 = metrics.f1_score(all_labels, all_predictions, average=average, zero_division=0)
        test_loss /= num_batches

        print(
             f"Test metrics: \n Accuracy: {float(acc):>6f}, F1 score: {float(f1):>6f}, Avg loss: {float(test_loss):>6f} \n"
        )

        return acc, f1


import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras.layers import (
    BatchNormalization,
    Conv1D,
    Dense,
    Dropout,
    GlobalAveragePooling1D,
    MaxPooling1D,
)
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

character_split_fn = lambda x: tf.strings.unicode_split(x, "UTF-8")
vectorize_layer = TextVectorization(output_mode="int", split=character_split_fn)

def get_basic_cnn_model_v0(num_classes, vocab_size):

    if num_classes == 2:
        last_layer = Dense(1)
        loss = tf.keras.losses.BinaryCrossentropy(from_logits=True)
        acc = tf.metrics.BinaryAccuracy(threshold=0.0)
        f1 = tfa.metrics.F1Score(num_classes=1, threshold=0.5, average="micro")
    else:
        last_layer = Dense(num_classes, activation="softmax")
        loss = 'categorical_crossentropy'
        acc = tf.metrics.CategoricalAccuracy()
        f1 = tfa.metrics.F1Score(num_classes=num_classes, average="micro")

    onehot_layer = tf.keras.layers.Lambda(lambda x: tf.one_hot(tf.cast(x, "int64"), vocab_size))

    model = tf.keras.Sequential(
        [
            onehot_layer,
            Conv1D(32, kernel_size=8, data_format="channels_last", activation="relu"),
            BatchNormalization(),
            MaxPooling1D(),
            Conv1D(16, kernel_size=8, data_format="channels_last", activation="relu"),
            BatchNormalization(),
            MaxPooling1D(),
            Conv1D(4, kernel_size=8, data_format="channels_last", activation="relu"),
            BatchNormalization(),
            MaxPooling1D(),
            Dropout(0.3),
            GlobalAveragePooling1D(),
            last_layer,
        ]
    )

    model.compile(
        loss=loss,
        optimizer='adam',
        metrics=[acc, f1]
    )

    return model




from Bio import SeqIO
from tqdm.autonotebook import tqdm


def fasta2loc(fasta_path, ref_dict, use_seq_ids=True):

    tree = {}
    nseqs = 0

    # building tree for seq searching
    for seq in SeqIO.parse(open(fasta_path, "r"), "fasta"):
        s = str(seq.seq)
        rev = str(seq.seq.reverse_complement())
        if use_seq_ids:
            sname = seq.name
        else:
            sname = s
        nseqs += 1

        _update_tree(tree, s, sname, "+")
        _update_tree(tree, rev, sname, "-")

    print(f"{nseqs} sequences read and parsed.")

    results = {}

    for chrom in tqdm(ref_dict):
        curr_positions = []
        # print(f"Processing chrom {chrom}.")

        for i, c in tqdm(enumerate(ref_dict[chrom]), total=len(ref_dict[chrom]), leave=False):

            prev_positions = curr_positions + [tree]
            curr_positions = []

            for pos in prev_positions:
                if c in pos:
                    pos = pos[c]
                    curr_positions.append(pos)

                    if "terminal" in pos:
                        results[pos["terminal"][0]] = (chrom, i - pos["terminal"][2] + 1, i + 1, pos["terminal"][1])

    print(f"{len(results.keys())} sequences found in the reference.")

    return results


def _update_tree(root, seq_str, seq_name, direction):
    # updates tree in `root` with a sequence `seq_str`
    position = root

    for c in seq_str:
        if c in position:
            position = position[c]
        else:
            position[c] = {}
            position = position[c]
    position["terminal"] = (seq_name, direction, len(seq_str))



import shutil
from pathlib import Path

import shutil
import gdown

# the files can be found in https://drive.google.com/drive/folders/1hV1swtSjun0TVYtuWBERkZVajvMgK-II (ask Petr for permission to write)
# the value is file_id that can be found by right-clicking on the file and selecting "Get link"
CLOUD_CACHE = {
    ("human_nontata_promoters", 0): "1VdUg0Zu8yfLS6QesBXwGz1PIQrTW3Ze4",
    ("demo_coding_vs_intergenomic_seqs", 0): "1cpXg0ULuTGF7h1_HTYvc6p8M-ee43t-v",
    ("demo_human_or_worm", 0): "1JW0-eTB-rJXvFcglqBo3pFZi1kyIWC3X",
    ("dummy_mouse_enhancers_ensembl", 0): "1dz7dwvyM1TVUsCyuxxxn1ZMMFV381sl2",
    ("human_enhancers_cohn", 0): "176563cDPQ5Y094WyoSBF02QjoVQhWuCh",
    ("human_enhancers_ensembl", 0): "1gZBEV_RGxJE8EON5OObdrp5Tp8JL0Fxb",
    ("human_ocr_ensembl", 0): "1y_LInRF2aRXysigpwv_oU3Q67VVxfk18",
    ("human_ensembl_regulatory", 0): "1GefcGAM-tklnmzfzZ9RC91tGDmOzrAXU"
}


def download_from_cloud_cache(file_key, dest_path, cloud_cache=CLOUD_CACHE, force_download=True):
    """
    Download a file from the cloud cache.
    """
    if file_key not in cloud_cache:
        raise ValueError(f"File ID {file_key} not in the cloud cache.")

    if force_download and Path(dest_path).exists():
        shutil.rmtree(str(dest_path))

    gdown.download(
        id = cloud_cache[file_key],
        output = str(dest_path) + ".zip",
    )

    shutil.unpack_archive(str(dest_path) + ".zip", Path(dest_path).parent)
    Path(str(dest_path) + ".zip").unlink()

    return Path(dest_path)



from genomic_benchmarks.data_check import is_downloaded
from genomic_benchmarks.loc2seq import download_dataset
from genomic_benchmarks.utils.paths import CACHE_PATH
from torch.utils.data import Dataset


class GenomicClfDataset(Dataset):
    """
    A class to represent generic genomic classification pytorch dataset.
    Instance of this class can be directly wrapped by pytorch DataLoader
    """

    def __init__(self, dset_name, split, force_download=False, version=None):
        """
        Parameters
            dset_name : str
                One of the existing dataset names, list available at TODO
            split : str
                train' or 'test'
            force_download : bool
                Whether to re-download already existing files
            version : int
                Version of the dataset
        """
        base_path = CACHE_PATH / dset_name
        if not is_downloaded(dset_name):
            download_dataset(dset_name, version=version, force_download=force_download)

        if split == "train" or split == "test":
            base_path = base_path / split
        else:
            raise Exception("Define split, train or test")

        self.all_paths = []
        self.all_labels = []
        label_mapper = {}

        for i, x in enumerate(base_path.iterdir()):
            label_mapper[x.stem] = i

        for label_type in label_mapper.keys():
            for x in (base_path / label_type).iterdir():
                self.all_paths.append(x)
                self.all_labels.append(label_mapper[label_type])

    def __len__(self):
        return len(self.all_paths)

    def __getitem__(self, idx):
        txt_path = self.all_paths[idx]
        with open(txt_path, "r") as f:
            content = f.read()
        x = content
        y = self.all_labels[idx]
        return x, y


def get_dataset(dataset_name, split, force_download=False, version=None):
    return GenomicClfDataset(dataset_name, split, force_download, version)


def DemoCodingVsIntergenomicSeqs(split, force_download=False, version=None):
    return GenomicClfDataset("demo_coding_vs_intergenomic_seqs", split, force_download, version)


def DemoHumanOrWorm(split, force_download=False, version=None):
    return GenomicClfDataset("demo_human_or_worm", split, force_download, version)


def DemoMouseEnhancers(split, force_download=False, version=None):
    return GenomicClfDataset("dummy_mouse_enhancers_ensembl", split, force_download, version)


def HumanEnhancersCohn(split, force_download=False, version=None):
    return GenomicClfDataset("human_enhancers_cohn", split, force_download, version)


def HumanEnhancersEnsembl(split, force_download=False, version=None):
    return GenomicClfDataset("human_enhancers_ensembl", split, force_download, version)


def HumanNontataPromoters(split, force_download=False, version=None):
    return GenomicClfDataset("human_nontata_promoters", split, force_download, version)

def HumanOcrEnsembl(split, force_download=False, version=None):
    return GenomicClfDataset("human_ocr_ensembl", split, force_download, version)



import types
from collections import Counter

import numpy as np
import torch
from torch.nn import ConstantPad1d
from torchtext.vocab import build_vocab_from_iterator, vocab

VARIABLE_LENGTH_DATASETS = [
    "human_enhancers_ensembl",
    "dummy_mouse_enhancers_ensembl",
    "human_ocr_ensembl",
    "human_ensembl_regulatory"
]


def coll_factory(vocab, tokenizer, device="cpu", pad_to_length=None):
    def coll(batch):
        xs, ys = [], []

        for text, label in batch:
            ys.append(torch.tensor([label], dtype=torch.float32))
            x = torch.tensor([vocab[token] for token in tokenizer(text)], dtype=torch.long)
            if pad_to_length != None:
                PAD_IDX = vocab["<pad>"]
                pad = ConstantPad1d((0, pad_to_length - len(x)), PAD_IDX)
                x = torch.tensor(pad(x), dtype=torch.long)
            xs.append(x)

        xs = torch.stack(xs)
        ys = torch.stack(ys)
        return xs.to(device), ys.to(device)

    return coll


#             torch.nn.functional.pad(
#                 torch.tensor([mapper[ch] for ch in text], dtype=torch.float32),
#                 (0,4707),
#                 mode='constant',
#                 value=4
#             )


class LetterTokenizer:
    def __init__(self, **kwargs):
        pass

    def __call__(self, items):
        if isinstance(items, str):
            return self.__tokenize_str(items)
        else:
            return (self.__tokenize_str(t) for t in items)

    def __tokenize_str(self, t):
        tokenized = list(t.replace("\n", ""))
        tokenized.append("<eos>")
        tokenized.insert(0, "<bos>")
        return tokenized


def build_vocab(dataset, tokenizer, use_padding):
    counter = Counter()
    for i in range(len(dataset)):
        counter.update(tokenizer(dataset[i][0]))
    #     print(counter.most_common())
    builded_voc = vocab(counter)
    if use_padding:
        builded_voc.append_token("<pad>")
    builded_voc.insert_token("<unk>", 0)
    builded_voc.set_default_index(0)
    return builded_voc


# todo: why build fn does not work as expected (iterator argument)
#     return build_vocab_from_iterator(
#         iterator = counter,
#         specials = ['<unk>', '<pad>', '<bos>', '<eos>'],
#         special_first = True)


def check_seq_lengths(dataset, use_padding):
    # Compute length of the longest sequence
    max_seq_len = max([len(dataset[i][0]) for i in range(len(dataset))])
    print("max_seq_len ", max_seq_len)
    same_length = [len(dataset[i][0]) == max_seq_len for i in range(len(dataset))]
    if not all(same_length):
        print("not all sequences are of the same length")

    # Count in tokens added in tokenizer '<bos>' and '<eos>' and the padding token <pad>
    if use_padding:
        len_with_tokens = max_seq_len + 3
    else:
        len_with_tokens = max_seq_len + 2
    return max_seq_len, len_with_tokens


def check_config(config):
    control_config = {
        "use_padding": bool,
        "run_on_gpu": bool,
        "dataset": str,
        "number_of_classes": int,
        "dataset_version": int,
        "force_download": bool,
        "epochs": int,
        "embedding_dim": int,
        "batch_size": int,
    }

    for key in config.keys():
        assert isinstance(config[key], control_config[key]), '"{}" in config should be of type {} but is {}'.format(
            key, control_config[key], type(config[key])
        )



from pathlib import Path

import numpy as np
import pandas as pd
from genomic_benchmarks.loc2seq.cloud_caching import CLOUD_CACHE
from genomic_benchmarks.utils.datasets import (
    _check_dataset_existence,
    _get_dataset_name,
    _guess_location,
)
from genomic_benchmarks.utils.paths import CACHE_PATH, DATASET_DIR_PATH

# TODO: Many of these functions are not prepared for the case when the folder in DATASET_DIR_PATH is not one benchmark but a set of benchmarks.


def info(interval_list_dataset, version=None, local_repo: bool = False):
    """
    Print info about the bechmark.

            Parameters:
                    interval_list_dataset (str or Path): Either a path or a name of dataset included in this package.

            Returns:
                    DataFrame with counts of seqeunces for each class in a training and testing sets.
    """

    interval_list_dataset = _guess_location(interval_list_dataset, local_repo)
    metadata = _check_dataset_existence(interval_list_dataset, version, local_repo)
    dataset_name = _get_dataset_name(interval_list_dataset)

    dfs = {}
    for c in metadata["classes"]:
        dfs[c] = {}
        for t in ["train", "test"]:
            dt_filename = interval_list_dataset / t / (c + ".csv.gz")
            dfs[c][t] = pd.read_csv(str(dt_filename), compression="gzip")

    classes = list(dfs.keys())
    print(f"Dataset `{dataset_name}` has {len(classes)} classes: " + ", ".join(classes) + ".\n")

    interval_lengths = np.concatenate([(dfs[c][t].end - dfs[c][t].start).to_numpy() for c in dfs for t in dfs[c]])

    if len(np.unique(interval_lengths)) == 1:
        print(f"All lengths of genomic intervals equals {interval_lengths[0]}.\n")
    else:
        print(
            f"The length of genomic intervals ranges from {np.min(interval_lengths)} to {np.max(interval_lengths)}, with average {np.mean(interval_lengths)} and median {np.median(interval_lengths)}.\n"
        )

    dfs_counts = {}
    for c in dfs:
        dfs_counts[c] = {}
        for t in dfs[c]:
            dfs_counts[c][t] = dfs[c][t].shape[0]
    dfs_counts_df = pd.DataFrame(dfs_counts).T

    print(
        f"Totally {dfs_counts_df.to_numpy().sum()} sequences have been found, {dfs_counts_df.train.sum()} for training and {dfs_counts_df.test.sum()} for testing."
    )

    return dfs_counts_df


def is_downloaded(interval_list_dataset, cache_path=CACHE_PATH):
    """
    Check if the dataset is downloaded.

            Parameters:
                    interval_list_dataset (str or Path): Either a path or a name of dataset included in this package.
                    cache_path (Path): Path to the cache directory.

            Returns:
                    bool: True if the dataset is downloaded, False otherwise.
    """

    dataset_name = _get_dataset_name(interval_list_dataset)
    cache_path = Path(cache_path) / dataset_name
    return cache_path.exists()


def list_datasets(dataset_path=DATASET_DIR_PATH, local_repo: bool = False):
    """
    List all interval datasets in the package.

            Parameters:
                    dataset_path (Path): Path to the datasets folder.

            Returns:
                    list: List of dataset names.
    """

    if local_repo:
        cache_path = Path(dataset_path)
        return [x.name for x in cache_path.iterdir() if x.is_dir()]
    else:
        return list({x[0] for x in CLOUD_CACHE})


# TODO: This ugly hack is deprecated and will be removed in the future.
def labels_in_order(dset_name):
    dir_path = CACHE_PATH / dset_name
    true_labels = []
    for i, label_path in enumerate(Path(dir_path / "test").iterdir()):
        for j in label_path.iterdir():
            true_labels.append(i)
    return true_labels
